import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;
import java.util.stream.Stream;

import org.graphstream.algorithm.Dijkstra;
import org.graphstream.algorithm.generator.Generator;
import org.graphstream.algorithm.generator.RandomGenerator;
import org.graphstream.graph.Edge;
import org.graphstream.graph.Graph;
import org.graphstream.graph.implementations.SingleGraph;
import org.graphstream.graph.Node;


public class ComparatifDijkstra {

    private static final int NB_TEST = 200;

    public static void main(String... args) {
        System.out.println("Lancement comparatif...");

        long[] resultatsCustom = new long[NB_TEST];
        long[] resultatsGraphStream = new long[NB_TEST];
        for (int i = 0; i < NB_TEST; i++) {
            SingleGraph graph = new SingleGraph("Comparatif graphe de taille identique");
            genererGraph(graph, 3000, 70, false);
            Node n = graph.getNode(0);

            resultatsCustom[i] = calculTempsCustom(graph, graph.getNode(0));;
            resultatsGraphStream[i] = calculTempsGraphStream(graph, graph.getNode(0));
        }
        System.out.println("Comparatif de " + NB_TEST  + " résultats pour un graphe de taille identique :");
        affichageComparatif(resultatsCustom, resultatsGraphStream);

        for (int i = 0; i < NB_TEST; i++)
        {
            SingleGraph graph = new SingleGraph("Comparatif graphe de taille identique");
            genererGraph(graph, (int)(Math.random() * 100) + 3000, (int)(Math.random() * 10) + 70, false);
            Node n = graph.getNode(0);

            resultatsCustom[i] = calculTempsCustom(graph, graph.getNode(0));;
            resultatsGraphStream[i] = calculTempsGraphStream(graph, graph.getNode(0));
        }

        System.out.println();
        System.out.println("Comparatif de " + NB_TEST  + " résultats pour un graphe de taille aléatoire :");
        affichageComparatif(resultatsCustom, resultatsGraphStream);
    }

    private static long calculTempsCustom(Graph g, Node n) {
        long time = System.currentTimeMillis();
        dijkstra(g, n, false);
        time = System.currentTimeMillis() - time;
        return time;
    }

    private static long calculTempsGraphStream(Graph g, Node n) {
        Dijkstra d = new Dijkstra();
        d.init(g);
        d.setSource(n);
        long time = System.currentTimeMillis();
        d.compute();
        time = System.currentTimeMillis() - time;
        return time;
    }

    private static void affichageComparatif(long[] resultatsCustom, long[] resultatsGraphStream) {
        System.out.println("Custom dijkstra :\n\ttemps :");
        System.out.println("\t\tMinimal: " + Arrays.stream(resultatsCustom).min().getAsLong() + "ms.");
        System.out.println("\t\tMaximal: " + Arrays.stream(resultatsCustom).max().getAsLong() + "ms.");
        System.out.println("\t\tMoyenne: " + Arrays.stream(resultatsCustom).average().getAsDouble() + "ms.");
        System.out.println("GraphStram dijkstra :\n\ttemps :");
        System.out.println("\t\tMinimal: " + Arrays.stream(resultatsGraphStream).min().getAsLong() + "ms.");
        System.out.println("\t\tMaximal: " + Arrays.stream(resultatsGraphStream).max().getAsLong() + "ms.");
        System.out.println("\t\tMoyenne: " + Arrays.stream(resultatsGraphStream).average().getAsDouble() + "ms.");
    }

    private static void genererGraph(Graph graph, int nbNoeuds, double degreMoyen, boolean estOriente) {

        RandomGenerator generator = new RandomGenerator(degreMoyen);
        generator.setDirectedEdges(estOriente, true);
        generator.addSink(graph);
        generator.addEdgeAttribute("poids");

        generator.begin();
        for (int i = 0; i < nbNoeuds; i++) {
            generator.nextEvents();
        }
        generator.end();

        graph.edges().forEach(e -> {
            double poids = (double) e.getAttribute("poids");
            e.setAttribute("poids", (int) (poids * 20)); //multiplie le nombre par celui que l'on veut
        });
    }




    public static void dijkstra(Graph graph, Node source, boolean oriente) {
        graph.nodes().forEach(n -> n.setAttribute("distance", -1));
        source.setAttribute("distance", 0);

        ArrayList<Node> listNodes = new ArrayList<>();
        listNodes.add(source);
        while (!listNodes.isEmpty()) {
            Node nCourante = listNodes.remove(0);

            // obtient toutes les arêtes restantes si le graphe est orienté sinon récupère les voisins restants.
            Stream<Node> nodes = oriente ? nCourante.leavingEdges().map(Edge::getTargetNode) : nCourante.neighborNodes();

            nodes.forEach(voisin -> {
                int distanceCourante = (int) nCourante.getAttribute("distance");
                int distanceVoisin = (int) voisin.getAttribute("distance");
                int poidsArrete = (int) voisin.getEdgeBetween(nCourante).getAttribute("poids");
                int distanceCalculee = distanceCourante + poidsArrete;

                // Si vrai une distance plus courte a été trouvé.
                if (distanceVoisin == -1 || distanceVoisin > distanceCalculee) {
                    voisin.setAttribute("distance", distanceCalculee);
                    voisin.setAttribute("parent", nCourante);
                    listNodes.add(voisin);
                }
            });
        }
    }
}